#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to check Hive partitions existence with Dataproc Metastore Sensor.

Note that Metastore service must be configured to use gRPC endpoints.
"""

from __future__ import annotations

import datetime
import os

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.google.cloud.hooks.gcs import _parse_gcs_url
from airflow.providers.google.cloud.operators.dataproc import (
    DataprocCreateClusterOperator,
    DataprocDeleteClusterOperator,
    DataprocSubmitJobOperator,
)
from airflow.providers.google.cloud.operators.dataproc_metastore import (
    DataprocMetastoreCreateServiceOperator,
    DataprocMetastoreDeleteServiceOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSDeleteBucketOperator
from airflow.providers.google.cloud.sensors.dataproc_metastore import MetastoreHivePartitionSensor
from airflow.providers.google.cloud.transfers.gcs_to_gcs import GCSToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "hive_partition_sensor"
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
REGION = "europe-west1"
NETWORK = "default"

METASTORE_SERVICE_ID = f"metastore-{DAG_ID}-{ENV_ID}".replace("_", "-")
METASTORE_TIMEOUT = 2400
METASTORE_SERVICE = {
    "name": METASTORE_SERVICE_ID,
    "hive_metastore_config": {
        "endpoint_protocol": "GRPC",
    },
    "network": f"projects/{PROJECT_ID}/global/networks/{NETWORK}",
}
METASTORE_SERVICE_QFN = f"projects/{PROJECT_ID}/locations/{REGION}/services/{METASTORE_SERVICE_ID}"
DATAPROC_CLUSTER_NAME = f"cluster-{DAG_ID}-{ENV_ID}".replace("_", "-")
DATAPROC_CLUSTER_CONFIG = {
    "master_config": {
        "num_instances": 1,
        "machine_type_uri": "n1-standard-2",
        "disk_config": {"boot_disk_type": "pd-standard", "boot_disk_size_gb": 1024},
    },
    "worker_config": {
        "num_instances": 2,
        "machine_type_uri": "n1-standard-2",
        "disk_config": {"boot_disk_type": "pd-standard", "boot_disk_size_gb": 1024},
    },
    "metastore_config": {
        "dataproc_metastore_service": METASTORE_SERVICE_QFN,
    },
    "gce_cluster_config": {
        "service_account_scopes": [
            "https://www.googleapis.com/auth/cloud-platform",
        ],
    },
}

TABLE_NAME = "transactions_partitioned"
COLUMN = "TransactionType"
PARTITION_1 = f"{COLUMN}=credit".lower()
PARTITION_2 = f"{COLUMN}=debit".lower()
SOURCE_DATA_BUCKET = "airflow-system-tests-resources"
SOURCE_DATA_PATH = "dataproc/hive"
SOURCE_DATA_FILE_NAME = "part-00000.parquet"
EXTERNAL_TABLE_BUCKET = "{{task_instance.xcom_pull(task_ids='get_hive_warehouse_bucket_task', key='bucket')}}"
QUERY_CREATE_EXTERNAL_TABLE = f"""
CREATE EXTERNAL TABLE IF NOT EXISTS transactions
(SubmissionDate DATE, TransactionAmount DOUBLE, TransactionType STRING)
STORED AS PARQUET
LOCATION 'gs://{EXTERNAL_TABLE_BUCKET}/{SOURCE_DATA_PATH}';
"""
QUERY_CREATE_PARTITIONED_TABLE = f"""
CREATE EXTERNAL TABLE IF NOT EXISTS {TABLE_NAME}
(SubmissionDate DATE, TransactionAmount DOUBLE)
PARTITIONED BY ({COLUMN} STRING);
"""
QUERY_COPY_DATA_WITH_PARTITIONS = f"""
SET hive.exec.dynamic.partition.mode=nonstrict;
INSERT INTO TABLE {TABLE_NAME} PARTITION ({COLUMN})
SELECT SubmissionDate,TransactionAmount,TransactionType FROM transactions;
"""

with DAG(
    DAG_ID,
    start_date=datetime.datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example", "dataproc", "metastore", "partition", "hive", "sensor"],
) as dag:
    create_metastore_service = DataprocMetastoreCreateServiceOperator(
        task_id="create_metastore_service",
        region=REGION,
        project_id=PROJECT_ID,
        service=METASTORE_SERVICE,
        service_id=METASTORE_SERVICE_ID,
        timeout=METASTORE_TIMEOUT,
    )

    create_cluster = DataprocCreateClusterOperator(
        task_id="create_cluster",
        cluster_name=DATAPROC_CLUSTER_NAME,
        project_id=PROJECT_ID,
        cluster_config=DATAPROC_CLUSTER_CONFIG,
        region=REGION,
    )

    @task(task_id="get_hive_warehouse_bucket_task")
    def get_hive_warehouse_bucket(**kwargs):
        """Return Hive Metastore Warehouse GCS bucket name."""
        ti = kwargs["ti"]
        metastore_service: dict = ti.xcom_pull(task_ids="create_metastore_service")
        config_overrides: dict = metastore_service["hive_metastore_config"]["config_overrides"]
        destination_dir: str = config_overrides["hive.metastore.warehouse.dir"]
        bucket, _ = _parse_gcs_url(destination_dir)
        ti.xcom_push(key="bucket", value=bucket)

    get_hive_warehouse_bucket_task = get_hive_warehouse_bucket()

    copy_source_data = GCSToGCSOperator(
        task_id="copy_source_data",
        source_bucket=SOURCE_DATA_BUCKET,
        source_object=f"{SOURCE_DATA_PATH}/{SOURCE_DATA_FILE_NAME}",
        destination_bucket=EXTERNAL_TABLE_BUCKET,
        destination_object=f"{SOURCE_DATA_PATH}/{SOURCE_DATA_FILE_NAME}",
    )

    create_external_table = DataprocSubmitJobOperator(
        task_id="create_external_table",
        job={
            "reference": {"project_id": PROJECT_ID},
            "placement": {"cluster_name": DATAPROC_CLUSTER_NAME},
            "hive_job": {"query_list": {"queries": [QUERY_CREATE_EXTERNAL_TABLE]}},
        },
        region=REGION,
        project_id=PROJECT_ID,
    )

    create_partitioned_table = DataprocSubmitJobOperator(
        task_id="create_partitioned_table",
        job={
            "reference": {"project_id": PROJECT_ID},
            "placement": {"cluster_name": DATAPROC_CLUSTER_NAME},
            "hive_job": {"query_list": {"queries": [QUERY_CREATE_PARTITIONED_TABLE]}},
        },
        region=REGION,
        project_id=PROJECT_ID,
    )

    partition_data = DataprocSubmitJobOperator(
        task_id="partition_data",
        job={
            "reference": {"project_id": PROJECT_ID},
            "placement": {"cluster_name": DATAPROC_CLUSTER_NAME},
            "hive_job": {"query_list": {"queries": [QUERY_COPY_DATA_WITH_PARTITIONS]}},
        },
        region=REGION,
        project_id=PROJECT_ID,
    )

    # [START how_to_cloud_dataproc_metastore_hive_partition_sensor]
    hive_partition_sensor = MetastoreHivePartitionSensor(
        task_id="hive_partition_sensor",
        service_id=METASTORE_SERVICE_ID,
        region=REGION,
        table=TABLE_NAME,
        partitions=[PARTITION_1, PARTITION_2],
    )
    # [END how_to_cloud_dataproc_metastore_hive_partition_sensor]

    delete_dataproc_cluster = DataprocDeleteClusterOperator(
        task_id="delete_dataproc_cluster",
        cluster_name=DATAPROC_CLUSTER_NAME,
        project_id=PROJECT_ID,
        region=REGION,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_metastore_service = DataprocMetastoreDeleteServiceOperator(
        task_id="delete_metastore_service",
        service_id=METASTORE_SERVICE_ID,
        project_id=PROJECT_ID,
        region=REGION,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_warehouse_bucket = GCSDeleteBucketOperator(
        task_id="delete_warehouse_bucket",
        bucket_name=EXTERNAL_TABLE_BUCKET,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_metastore_service
        >> create_cluster
        >> get_hive_warehouse_bucket_task
        >> copy_source_data
        >> create_external_table
        >> create_partitioned_table
        # TEST BODY
        >> partition_data
        >> hive_partition_sensor
        # TEST TEARDOWN
        >> [delete_dataproc_cluster, delete_metastore_service, delete_warehouse_bucket]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that translates text in Google Cloud Translate using V3 API version
service in the Google Cloud.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.translate import (
    TranslateCreateGlossaryOperator,
    TranslateDeleteGlossaryOperator,
    TranslateListGlossariesOperator,
    TranslateUpdateGlossaryOperator,
)

DAG_ID = "translate_glossary"
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
REGION = "us-central1"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"

GLOSSARY_FILE_INPUT = {
    "gcs_source": {"input_uri": f"gs://{RESOURCE_DATA_BUCKET}/V3_translate/glossaries/glossary_sample.tsv"}
}
UPDATE_GLOSSARY_FILE_INPUT = {
    "gcs_source": {
        "input_uri": f"gs://{RESOURCE_DATA_BUCKET}/V3_translate/glossaries/glossary_update_sample.tsv"
    }
}


with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2024, 1, 1),
    catchup=False,
    tags=["example", "glossary_translate", "translate_V3"],
) as dag:
    # [START howto_operator_translate_create_glossary]
    create_glossary = TranslateCreateGlossaryOperator(
        task_id="glossary_create",
        project_id=PROJECT_ID,
        location=REGION,
        input_config=GLOSSARY_FILE_INPUT,
        glossary_id=f"glossary_new_{PROJECT_ID}_{ENV_ID}",
        language_pair={"source_language_code": "en", "target_language_code": "es"},
    )
    # [END howto_operator_translate_create_glossary]

    # [START howto_operator_translate_update_glossary]
    glossary_id = create_glossary.output["glossary_id"]
    update_glossary = TranslateUpdateGlossaryOperator(
        task_id="glossary_update",
        project_id=PROJECT_ID,
        location=REGION,
        new_input_config=UPDATE_GLOSSARY_FILE_INPUT,
        new_display_name=f"gl_{PROJECT_ID}_updated",
        glossary_id=glossary_id,
    )
    # [END howto_operator_translate_update_glossary]

    # [START howto_operator_translate_list_glossaries]
    list_glossaries = TranslateListGlossariesOperator(
        task_id="list_glossaries",
        page_size=100,
        project_id=PROJECT_ID,
        location=REGION,
    )
    # [END howto_operator_translate_list_glossaries]

    # [START howto_operator_translate_delete_glossary]
    delete_glossary = TranslateDeleteGlossaryOperator(
        task_id="delete_glossary",
        glossary_id=glossary_id,
        project_id=PROJECT_ID,
        location=REGION,
    )
    # [END howto_operator_translate_delete_glossary]

    (
        # TEST BODY
        create_glossary >> update_glossary >> list_glossaries >> delete_glossary
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import call, patch

import pytest

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.datastore import DatastoreHook

GCP_PROJECT_ID = "test"


def mock_init(
    self,
    gcp_conn_id,
    impersonation_chain=None,
):
    pass


class TestDatastoreHook:
    def setup_method(self):
        with patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__", new=mock_init
        ):
            self.datastore_hook = DatastoreHook()

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook._authorize")
    @patch("airflow.providers.google.cloud.hooks.datastore.build")
    def test_get_conn(self, mock_build, mock_authorize):
        conn = self.datastore_hook.get_conn()

        mock_build.assert_called_once_with(
            "datastore", "v1", http=mock_authorize.return_value, cache_discovery=False
        )
        assert conn == mock_build.return_value
        assert conn == self.datastore_hook.connection

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_allocate_ids(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        partial_keys = []

        keys = self.datastore_hook.allocate_ids(partial_keys=partial_keys, project_id=GCP_PROJECT_ID)

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        allocate_ids = projects.return_value.allocateIds
        allocate_ids.assert_called_once_with(projectId=GCP_PROJECT_ID, body={"keys": partial_keys})
        execute = allocate_ids.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert keys == execute.return_value["keys"]

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_allocate_ids_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.connection = mock_get_conn.return_value
        partial_keys = []

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.allocate_ids(partial_keys=partial_keys)
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_begin_transaction(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value

        transaction = self.datastore_hook.begin_transaction(
            project_id=GCP_PROJECT_ID,
            transaction_options={},
        )

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        begin_transaction = projects.return_value.beginTransaction
        begin_transaction.assert_called_once_with(projectId=GCP_PROJECT_ID, body={"transactionOptions": {}})
        execute = begin_transaction.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert transaction == execute.return_value["transaction"]

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_begin_transaction_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.connection = mock_get_conn.return_value
        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.begin_transaction()
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_commit(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        body = {"item": "a"}

        resp = self.datastore_hook.commit(body=body, project_id=GCP_PROJECT_ID)

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        commit = projects.return_value.commit
        commit.assert_called_once_with(projectId=GCP_PROJECT_ID, body=body)
        execute = commit.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_commit_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.connection = mock_get_conn.return_value
        body = {"item": "a"}

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.commit(body=body)
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_lookup(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        keys = []
        read_consistency = "ENUM"
        transaction = "transaction"

        resp = self.datastore_hook.lookup(
            keys=keys, read_consistency=read_consistency, transaction=transaction, project_id=GCP_PROJECT_ID
        )

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        lookup = projects.return_value.lookup
        lookup.assert_called_once_with(
            projectId=GCP_PROJECT_ID,
            body={"keys": keys, "readConsistency": read_consistency, "transaction": transaction},
        )
        execute = lookup.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_lookup_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.connection = mock_get_conn.return_value
        keys = []
        read_consistency = "ENUM"
        transaction = "transaction"

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.lookup(
                keys=keys,
                read_consistency=read_consistency,
                transaction=transaction,
            )
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_rollback(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        transaction = "transaction"

        self.datastore_hook.rollback(transaction=transaction, project_id=GCP_PROJECT_ID)

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        rollback = projects.return_value.rollback
        rollback.assert_called_once_with(projectId=GCP_PROJECT_ID, body={"transaction": transaction})
        execute = rollback.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_rollback_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.connection = mock_get_conn.return_value
        transaction = "transaction"

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.rollback(transaction=transaction)
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_run_query(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        body = {"item": "a"}

        resp = self.datastore_hook.run_query(body=body, project_id=GCP_PROJECT_ID)

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        run_query = projects.return_value.runQuery
        run_query.assert_called_once_with(projectId=GCP_PROJECT_ID, body=body)
        execute = run_query.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value["batch"]

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_run_query_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.connection = mock_get_conn.return_value
        body = {"item": "a"}

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.run_query(body=body)
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_get_operation(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        name = "name"

        resp = self.datastore_hook.get_operation(name=name)

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        operations = projects.return_value.operations
        operations.assert_called_once_with()
        get = operations.return_value.get
        get.assert_called_once_with(name=name)
        execute = get.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_delete_operation(self, mock_get_conn):
        self.datastore_hook.connection = mock_get_conn.return_value
        name = "name"

        resp = self.datastore_hook.delete_operation(name=name)

        projects = self.datastore_hook.connection.projects
        projects.assert_called_once_with()
        operations = projects.return_value.operations
        operations.assert_called_once_with()
        delete = operations.return_value.delete
        delete.assert_called_once_with(name=name)
        execute = delete.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value

    @patch("airflow.providers.google.cloud.hooks.datastore.time.sleep")
    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_operation",
        side_effect=[
            {"metadata": {"common": {"state": "PROCESSING"}}},
            {"metadata": {"common": {"state": "NOT PROCESSING"}}},
        ],
    )
    def test_poll_operation_until_done(self, mock_get_operation, mock_time_sleep):
        name = "name"
        polling_interval_in_seconds = 10

        result = self.datastore_hook.poll_operation_until_done(name, polling_interval_in_seconds)

        mock_get_operation.assert_has_calls([call(name), call(name)])
        mock_time_sleep.assert_called_once_with(polling_interval_in_seconds)
        assert result == {"metadata": {"common": {"state": "NOT PROCESSING"}}}

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_export_to_storage_bucket(self, mock_get_conn):
        self.datastore_hook.admin_connection = mock_get_conn.return_value
        bucket = "bucket"
        namespace = None
        entity_filter = {}
        labels = {}

        resp = self.datastore_hook.export_to_storage_bucket(
            bucket=bucket,
            namespace=namespace,
            entity_filter=entity_filter,
            labels=labels,
            project_id=GCP_PROJECT_ID,
        )

        projects = self.datastore_hook.admin_connection.projects
        projects.assert_called_once_with()
        export = projects.return_value.export
        export.assert_called_once_with(
            projectId=GCP_PROJECT_ID,
            body={
                "outputUrlPrefix": "gs://" + "/".join(filter(None, [bucket, namespace])),
                "entityFilter": entity_filter,
                "labels": labels,
            },
        )
        execute = export.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_export_to_storage_bucket_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.admin_connection = mock_get_conn.return_value
        bucket = "bucket"
        namespace = None
        entity_filter = {}
        labels = {}

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.export_to_storage_bucket(
                bucket=bucket,
                namespace=namespace,
                entity_filter=entity_filter,
                labels=labels,
            )
        assert "project_id" in str(ctx.value)

    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_import_from_storage_bucket(self, mock_get_conn):
        self.datastore_hook.admin_connection = mock_get_conn.return_value
        bucket = "bucket"
        file = "file"
        namespace = None
        entity_filter = {}
        labels = {}

        resp = self.datastore_hook.import_from_storage_bucket(
            bucket=bucket,
            file=file,
            namespace=namespace,
            entity_filter=entity_filter,
            labels=labels,
            project_id=GCP_PROJECT_ID,
        )

        projects = self.datastore_hook.admin_connection.projects
        projects.assert_called_once_with()
        import_ = projects.return_value.import_
        import_.assert_called_once_with(
            projectId=GCP_PROJECT_ID,
            body={
                "inputUrl": "gs://" + "/".join(filter(None, [bucket, namespace, file])),
                "entityFilter": entity_filter,
                "labels": labels,
            },
        )
        execute = import_.return_value.execute
        execute.assert_called_once_with(num_retries=mock.ANY)
        assert resp == execute.return_value

    @patch(
        "airflow.providers.google.cloud.hooks.datastore.DatastoreHook.project_id",
        new_callable=mock.PropertyMock,
        return_value=None,
    )
    @patch("airflow.providers.google.cloud.hooks.datastore.DatastoreHook.get_conn")
    def test_import_from_storage_bucket_no_project_id(self, mock_get_conn, mock_project_id):
        self.datastore_hook.admin_connection = mock_get_conn.return_value
        bucket = "bucket"
        file = "file"
        namespace = None
        entity_filter = {}
        labels = {}

        with pytest.raises(AirflowException) as ctx:
            self.datastore_hook.import_from_storage_bucket(
                bucket=bucket,
                file=file,
                namespace=namespace,
                entity_filter=entity_filter,
                labels=labels,
            )
        assert "project_id" in str(ctx.value)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Sequence
from enum import Enum
from typing import TYPE_CHECKING, Any

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.cloud_run import CloudRunAsyncHook
from airflow.triggers.base import BaseTrigger, TriggerEvent

if TYPE_CHECKING:
    from google.longrunning import operations_pb2

DEFAULT_BATCH_LOCATION = "us-central1"


class RunJobStatus(Enum):
    """Enum to represent the status of a job run."""

    SUCCESS = "Success"
    FAIL = "Fail"
    TIMEOUT = "Timeout"


class CloudRunJobFinishedTrigger(BaseTrigger):
    """
    Cloud Run trigger to check if templated job has been finished.

    :param operation_name: Required. Name of the operation.
    :param job_name: Required. Name of the job.
    :param project_id: Required. the Google Cloud project ID in which the job was started.
    :param location: Optional. the location where job is executed.
        If set to None then the value of DEFAULT_BATCH_LOCATION will be used.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token of the last account
        in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param poll_sleep: Polling period in seconds to check for the status.
    :timeout: The time to wait before failing the operation.
    """

    def __init__(
        self,
        operation_name: str,
        job_name: str,
        project_id: str | None,
        location: str = DEFAULT_BATCH_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        polling_period_seconds: float = 10,
        timeout: float | None = None,
    ):
        super().__init__()
        self.project_id = project_id
        self.job_name = job_name
        self.operation_name = operation_name
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.polling_period_seconds = polling_period_seconds
        self.timeout = timeout
        self.impersonation_chain = impersonation_chain

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.cloud_run.CloudRunJobFinishedTrigger",
            {
                "project_id": self.project_id,
                "operation_name": self.operation_name,
                "job_name": self.job_name,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "polling_period_seconds": self.polling_period_seconds,
                "timeout": self.timeout,
                "impersonation_chain": self.impersonation_chain,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        timeout = self.timeout
        hook = self._get_async_hook()
        while timeout is None or timeout > 0:
            operation: operations_pb2.Operation = await hook.get_operation(self.operation_name)
            if operation.done:
                # An operation can only have one of those two combinations: if it is failed, then
                # the error field will be populated, else, then the response field will be.
                if operation.error.SerializeToString():
                    yield TriggerEvent(
                        {
                            "status": RunJobStatus.FAIL.value,
                            "operation_error_code": operation.error.code,
                            "operation_error_message": operation.error.message,
                            "job_name": self.job_name,
                        }
                    )
                else:
                    yield TriggerEvent(
                        {
                            "status": RunJobStatus.SUCCESS.value,
                            "job_name": self.job_name,
                        }
                    )
                return
            elif operation.error.message:
                raise AirflowException(f"Cloud Run Job error: {operation.error.message}")

            if timeout is not None:
                timeout -= self.polling_period_seconds

            if timeout is None or timeout > 0:
                await asyncio.sleep(self.polling_period_seconds)

        yield TriggerEvent(
            {
                "status": RunJobStatus.TIMEOUT.value,
                "job_name": self.job_name,
            }
        )

    def _get_async_hook(self) -> CloudRunAsyncHook:
        return CloudRunAsyncHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import tempfile
from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator

try:
    from airflow.providers.microsoft.azure.hooks.wasb import WasbHook
except ModuleNotFoundError as e:
    from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

    raise AirflowOptionalProviderFeatureException(e)

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class AzureBlobStorageToGCSOperator(BaseOperator):
    """
    Operator transfers data from Azure Blob Storage to specified bucket in Google Cloud Storage.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AzureBlobStorageToGCSOperator`

    :param wasb_conn_id: Reference to the wasb connection.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param blob_name: Name of the blob
    :param container_name: Name of the container
    :param bucket_name: The bucket to upload to
    :param object_name: The object name to set when uploading the file
    :param filename: The local file path to the file to be uploaded
    :param gzip: Option to compress local file or file data for upload
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    def __init__(
        self,
        *,
        wasb_conn_id="wasb_default",
        gcp_conn_id: str = "google_cloud_default",
        blob_name: str,
        container_name: str,
        bucket_name: str,
        object_name: str,
        filename: str,
        gzip: bool,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.wasb_conn_id = wasb_conn_id
        self.gcp_conn_id = gcp_conn_id
        self.blob_name = blob_name
        self.container_name = container_name
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.filename = filename
        self.gzip = gzip
        self.impersonation_chain = impersonation_chain

    template_fields: Sequence[str] = (
        "blob_name",
        "container_name",
        "bucket_name",
        "object_name",
        "filename",
    )

    def execute(self, context: Context) -> str:
        azure_hook = WasbHook(wasb_conn_id=self.wasb_conn_id)
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        with tempfile.NamedTemporaryFile() as temp_file:
            self.log.info("Downloading data from blob: %s", self.blob_name)
            azure_hook.get_file(
                file_path=temp_file.name,
                container_name=self.container_name,
                blob_name=self.blob_name,
            )
            self.log.info(
                "Uploading data from blob's: %s into GCP bucket: %s", self.object_name, self.bucket_name
            )
            gcs_hook.upload(
                bucket_name=self.bucket_name,
                object_name=self.object_name,
                filename=temp_file.name,
                gzip=self.gzip,
            )
            self.log.info(
                "Resources have been uploaded from blob: %s to GCS bucket:%s",
                self.blob_name,
                self.bucket_name,
            )
        return f"gs://{self.bucket_name}/{self.object_name}"

    def get_openlineage_facets_on_start(self):
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.openlineage.extractors import OperatorLineage

        wasb_hook = WasbHook(wasb_conn_id=self.wasb_conn_id)
        account_name = wasb_hook.get_conn().account_name

        return OperatorLineage(
            inputs=[Dataset(namespace=f"wasbs://{self.container_name}@{account_name}", name=self.blob_name)],
            outputs=[Dataset(namespace=f"gs://{self.bucket_name}", name=self.object_name)],
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains Google PubSub operators.

.. spelling:word-list::

    MessageStoragePolicy
"""

from __future__ import annotations

from collections.abc import Callable, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.pubsub_v1.types import (
    DeadLetterPolicy,
    Duration,
    ExpirationPolicy,
    MessageStoragePolicy,
    PushConfig,
    ReceivedMessage,
    RetryPolicy,
    SchemaSettings,
)

from airflow.providers.common.compat.sdk import AirflowException, conf
from airflow.providers.google.cloud.hooks.pubsub import PubSubHook
from airflow.providers.google.cloud.links.pubsub import PubSubSubscriptionLink, PubSubTopicLink
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.cloud.triggers.pubsub import PubsubPullTrigger
from airflow.providers.google.common.consts import GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from google.api_core.retry import Retry

    from airflow.providers.common.compat.sdk import Context
    from airflow.providers.openlineage.extractors import OperatorLineage


class PubSubCreateTopicOperator(GoogleCloudBaseOperator):
    """
    Create a PubSub topic.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:PubSubCreateTopicOperator`

    By default, if the topic already exists, this operator will
    not cause the DAG to fail. ::

        with DAG("successful DAG") as dag:
            create_topic = PubSubCreateTopicOperator(project_id="my-project", topic="my_new_topic")
            create_topic_again = PubSubCreateTopicOperator(project_id="my-project", topic="my_new_topic")

            create_topic >> create_topic_again

    The operator can be configured to fail if the topic already exists. ::

        with DAG("failing DAG") as dag:
            create_topic = PubSubCreateTopicOperator(project_id="my-project", topic="my_new_topic")
            create_topic_again = PubSubCreateTopicOperator(
                project_id="my-project", topic="my_new_topic", fail_if_exists=True
            )

            create_topic >> create_topic_again

    Both ``project_id`` and ``topic`` are templated so you can use Jinja templating in their values.

    :param project_id: Optional, the Google Cloud project ID where the topic will be created.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param topic: the topic to create. Do not include the
        full topic path. In other words, instead of
        ``projects/{project}/topics/{topic}``, provide only
        ``{topic}``. (templated)
    :param gcp_conn_id: The connection ID to use connecting to
        Google Cloud.
    :param labels: Client-assigned labels; see
        https://cloud.google.com/pubsub/docs/labels
    :param message_storage_policy: Policy constraining the set
        of Google Cloud regions where messages published to
        the topic may be stored. If not present, then no constraints
        are in effect.
        Union[dict, google.cloud.pubsub_v1.types.MessageStoragePolicy]
    :param kms_key_name: The resource name of the Cloud KMS CryptoKey
        to be used to protect access to messages published on this topic.
        The expected format is
        ``projects/*/locations/*/keyRings/*/cryptoKeys/*``.
    :param retry: (Optional) A retry object used to retry requests.
        If None is specified, requests will not be retried.
    :param timeout: (Optional) The amount of time, in seconds, to wait for the request
        to complete. Note that if retry is specified, the timeout applies to each
        individual attempt.
    :param metadata: (Optional) Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "topic",
        "impersonation_chain",
    )
    ui_color = "#0273d4"
    operator_extra_links = (PubSubTopicLink(),)

    def __init__(
        self,
        *,
        topic: str,
        project_id: str = PROVIDE_PROJECT_ID,
        fail_if_exists: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        labels: dict[str, str] | None = None,
        message_storage_policy: dict | MessageStoragePolicy = None,
        kms_key_name: str | None = None,
        schema_settings: dict | SchemaSettings = None,
        message_retention_duration: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.topic = topic
        self.fail_if_exists = fail_if_exists
        self.gcp_conn_id = gcp_conn_id
        self.labels = labels
        self.message_storage_policy = message_storage_policy
        self.kms_key_name = kms_key_name
        self.schema_settings = schema_settings
        self.message_retention_duration = message_retention_duration
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = PubSubHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Creating topic %s", self.topic)
        hook.create_topic(
            project_id=self.project_id,
            topic=self.topic,
            fail_if_exists=self.fail_if_exists,
            labels=self.labels,
            message_storage_policy=self.message_storage_policy,
            kms_key_name=self.kms_key_name,
            schema_settings=self.schema_settings,
            message_retention_duration=self.message_retention_duration,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        self.log.info("Created topic %s", self.topic)
        PubSubTopicLink.persist(
            context=context,
            topic_id=self.topic,
            project_id=self.project_id or hook.project_id,
        )


class PubSubCreateSubscriptionOperator(GoogleCloudBaseOperator):
    """
    Create a PubSub subscription.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:PubSubCreateSubscriptionOperator`

    By default, the subscription will be created in ``project_id``. If
    ``subscription_project_id`` is specified and the Google Cloud credentials allow, the
    Subscription can be created in a different project from its topic.

    By default, if the subscription already exists, this operator will
    not cause the DAG to fail. However, the topic must exist in the project. ::

        with DAG("successful DAG") as dag:
            create_subscription = PubSubCreateSubscriptionOperator(
                project_id="my-project", topic="my-topic", subscription="my-subscription"
            )
            create_subscription_again = PubSubCreateSubscriptionOperator(
                project_id="my-project", topic="my-topic", subscription="my-subscription"
            )

            create_subscription >> create_subscription_again


    The operator can be configured to fail if the subscription already exists.
    ::

        with DAG("failing DAG") as dag:
            create_subscription = PubSubCreateSubscriptionOperator(
                project_id="my-project", topic="my-topic", subscription="my-subscription"
            )
            create_subscription_again = PubSubCreateSubscriptionOperator(
                project_id="my-project", topic="my-topic", subscription="my-subscription", fail_if_exists=True
            )

            create_subscription >> create_subscription_again

    Finally, subscription is not required. If not passed, the operator will
    generated a universally unique identifier for the subscription's name. ::

        with DAG("DAG") as dag:
            PubSubCreateSubscriptionOperator(project_id="my-project", topic="my-topic")

    ``project_id``, ``topic``, ``subscription``, ``subscription_project_id`` and
    ``impersonation_chain`` are templated so you can use Jinja templating in their values.

    :param project_id: Optional, the Google Cloud project ID where the topic exists.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param topic: the topic to create. Do not include the
        full topic path. In other words, instead of
        ``projects/{project}/topics/{topic}``, provide only
        ``{topic}``. (templated)
    :param subscription: the Pub/Sub subscription name. If empty, a random
        name will be generated using the uuid module
    :param subscription_project_id: the Google Cloud project ID where the subscription
        will be created. If empty, ``topic_project`` will be used.
    :param ack_deadline_secs: Number of seconds that a subscriber has to
        acknowledge each message pulled from the subscription
    :param gcp_conn_id: The connection ID to use connecting to
        Google Cloud.
    :param push_config: If push delivery is used with this subscription,
        this field is used to configure it. An empty ``pushConfig`` signifies
        that the subscriber will pull and ack messages using API methods.
    :param retain_acked_messages: Indicates whether to retain acknowledged
        messages. If true, then messages are not expunged from the subscription's
        backlog, even if they are acknowledged, until they fall out of the
        ``message_retention_duration`` window. This must be true if you would
        like to Seek to a timestamp.
    :param message_retention_duration: How long to retain unacknowledged messages
        in the subscription's backlog, from the moment a message is published. If
        ``retain_acked_messages`` is true, then this also configures the
        retention of acknowledged messages, and thus configures how far back in
        time a ``Seek`` can be done. Defaults to 7 days. Cannot be more than 7
        days or less than 10 minutes.
    :param labels: Client-assigned labels; see
        https://cloud.google.com/pubsub/docs/labels
    :param enable_message_ordering: If true, messages published with the same
        ordering_key in PubsubMessage will be delivered to the subscribers in the order
        in which they are received by the Pub/Sub system. Otherwise, they may be
        delivered in any order.
    :param expiration_policy: A policy that specifies the conditions for this
        subscription's expiration. A subscription is considered active as long as any
        connected subscriber is successfully consuming messages from the subscription or
        is issuing operations on the subscription. If expiration_policy is not set,
        a default policy with ttl of 31 days will be used. The minimum allowed value for
        expiration_policy.ttl is 1 day.
    :param filter_: An expression written in the Cloud Pub/Sub filter language. If
        non-empty, then only PubsubMessages whose attributes field matches the filter are
        delivered on this subscription. If empty, then no messages are filtered out.
    :param dead_letter_policy: A policy that specifies the conditions for dead lettering
        messages in this subscription. If dead_letter_policy is not set, dead lettering is
        disabled.
    :param retry_policy: A policy that specifies how Pub/Sub retries message delivery
        for this subscription. If not set, the default retry policy is applied. This
        generally implies that messages will be retried as soon as possible for healthy
        subscribers. RetryPolicy will be triggered on NACKs or acknowledgement deadline
        exceeded events for a given message.
    :param retry: (Optional) A retry object used to retry requests.
        If None is specified, requests will not be retried.
    :param timeout: (Optional) The amount of time, in seconds, to wait for the request
        to complete. Note that if retry is specified, the timeout applies to each
        individual attempt.
    :param metadata: (Optional) Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "topic",
        "subscription",
        "subscription_project_id",
        "impersonation_chain",
    )
    ui_color = "#0273d4"
    operator_extra_links = (PubSubSubscriptionLink(),)

    def __init__(
        self,
        *,
        topic: str,
        project_id: str = PROVIDE_PROJECT_ID,
        subscription: str | None = None,
        subscription_project_id: str | None = None,
        ack_deadline_secs: int = 10,
        fail_if_exists: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        push_config: dict | PushConfig | None = None,
        retain_acked_messages: bool | None = None,
        message_retention_duration: dict | Duration | None = None,
        labels: dict[str, str] | None = None,
        enable_message_ordering: bool = False,
        expiration_policy: dict | ExpirationPolicy | None = None,
        filter_: str | None = None,
        dead_letter_policy: dict | DeadLetterPolicy | None = None,
        retry_policy: dict | RetryPolicy | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.topic = topic
        self.subscription = subscription
        self.subscription_project_id = subscription_project_id
        self.ack_deadline_secs = ack_deadline_secs
        self.fail_if_exists = fail_if_exists
        self.gcp_conn_id = gcp_conn_id
        self.push_config = push_config
        self.retain_acked_messages = retain_acked_messages
        self.message_retention_duration = message_retention_duration
        self.labels = labels
        self.enable_message_ordering = enable_message_ordering
        self.expiration_policy = expiration_policy
        self.filter_ = filter_
        self.dead_letter_policy = dead_letter_policy
        self.retry_policy = retry_policy
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.impersonation_chain = impersonation_chain
        self._resolved_subscription_name: str | None = None

    @cached_property
    def pubsub_hook(self):
        return PubSubHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

    def execute(self, context: Context) -> str:
        self.log.info("Creating subscription for topic %s", self.topic)
        result = self.pubsub_hook.create_subscription(
            project_id=self.project_id,
            topic=self.topic,
            subscription=self.subscription,
            subscription_project_id=self.subscription_project_id,
            ack_deadline_secs=self.ack_deadline_secs,
            fail_if_exists=self.fail_if_exists,
            push_config=self.push_config,
            retain_acked_messages=self.retain_acked_messages,
            message_retention_duration=self.message_retention_duration,
            labels=self.labels,
            enable_message_ordering=self.enable_message_ordering,
            expiration_policy=self.expiration_policy,
            filter_=self.filter_,
            dead_letter_policy=self.dead_letter_policy,
            retry_policy=self.retry_policy,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        self.log.info("Created subscription for topic %s", self.topic)

        # Store resolved subscription for Open Lineage
        self._resolved_subscription_name = self.subscription or result

        PubSubSubscriptionLink.persist(
            context=context,
            subscription_id=self._resolved_subscription_name,  # result returns subscription name
            project_id=self.project_id or self.pubsub_hook.project_id,
        )
        return result

    def get_openlineage_facets_on_complete(self, _) -> OperatorLineage:
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.openlineage.extractors import OperatorLineage

        topic_project_id = self.project_id or self.pubsub_hook.project_id
        subscription_project_id = self.subscription_project_id or topic_project_id

        return OperatorLineage(
            inputs=[Dataset(namespace="pubsub", name=f"topic:{topic_project_id}:{self.topic}")],
            outputs=[
                Dataset(
                    namespace="pubsub",
                    name=f"subscription:{subscription_project_id}:{self._resolved_subscription_name}",
                )
            ],
        )


class PubSubDeleteTopicOperator(GoogleCloudBaseOperator):
    """
    Delete a PubSub topic.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:PubSubDeleteTopicOperator`

    By default, if the topic does not exist, this operator will
    not cause the DAG to fail. ::

        with DAG("successful DAG") as dag:
            PubSubDeleteTopicOperator(project_id="my-project", topic="non_existing_topic")

    The operator can be configured to fail if the topic does not exist. ::

        with DAG("failing DAG") as dag:
            PubSubDeleteTopicOperator(
                project_id="my-project",
                topic="non_existing_topic",
                fail_if_not_exists=True,
            )

    Both ``project_id`` and ``topic`` are templated so you can use Jinja templating in their values.

    :param project_id: Optional, the Google Cloud project ID in which to work (templated).
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param topic: the topic to delete. Do not include the
        full topic path. In other words, instead of
        ``projects/{project}/topics/{topic}``, provide only
        ``{topic}``. (templated)
    :param fail_if_not_exists: If True and the topic does not exist, fail
        the task
    :param gcp_conn_id: The connection ID to use connecting to
        Google Cloud.
    :param retry: (Optional) A retry object used to retry requests.
        If None is specified, requests will not be retried.
    :param timeout: (Optional) The amount of time, in seconds, to wait for the request
        to complete. Note that if retry is specified, the timeout applies to each
        individual attempt.
    :param metadata: (Optional) Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "topic",
        "impersonation_chain",
    )
    ui_color = "#cb4335"

    def __init__(
        self,
        *,
        topic: str,
        project_id: str = PROVIDE_PROJECT_ID,
        fail_if_not_exists: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.topic = topic
        self.fail_if_not_exists = fail_if_not_exists
        self.gcp_conn_id = gcp_conn_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = PubSubHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Deleting topic %s", self.topic)
        hook.delete_topic(
            project_id=self.project_id,
            topic=self.topic,
            fail_if_not_exists=self.fail_if_not_exists,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        self.log.info("Deleted topic %s", self.topic)


class PubSubDeleteSubscriptionOperator(GoogleCloudBaseOperator):
    """
    Delete a PubSub subscription.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:PubSubDeleteSubscriptionOperator`

    By default, if the subscription does not exist, this operator will
    not cause the DAG to fail. ::

        with DAG("successful DAG") as dag:
            PubSubDeleteSubscriptionOperator(project_id="my-project", subscription="non-existing")

    The operator can be configured to fail if the subscription already exists.

    ::

        with DAG("failing DAG") as dag:
            PubSubDeleteSubscriptionOperator(
                project_id="my-project",
                subscription="non-existing",
                fail_if_not_exists=True,
            )

    ``project_id``, and ``subscription`` are templated so you can use Jinja templating in their values.

    :param project_id: Optional, the Google Cloud project ID in which to work (templated).
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param subscription: the subscription to delete. Do not include the
        full subscription path. In other words, instead of
        ``projects/{project}/subscription/{subscription}``, provide only
        ``{subscription}``. (templated)
    :param fail_if_not_exists: If True and the subscription does not exist,
        fail the task
    :param gcp_conn_id: The connection ID to use connecting to
        Google Cloud.
    :param retry: (Optional) A retry object used to retry requests.
        If None is specified, requests will not be retried.
    :param timeout: (Optional) The amount of time, in seconds, to wait for the request
        to complete. Note that if retry is specified, the timeout applies to each
        individual attempt.
    :param metadata: (Optional) Additional metadata that is provided to the method.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "subscription",
        "impersonation_chain",
    )
    ui_color = "#cb4335"

    def __init__(
        self,
        *,
        subscription: str,
        project_id: str = PROVIDE_PROJECT_ID,
        fail_if_not_exists: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.subscription = subscription
        self.fail_if_not_exists = fail_if_not_exists
        self.gcp_conn_id = gcp_conn_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = PubSubHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Deleting subscription %s", self.subscription)
        hook.delete_subscription(
            project_id=self.project_id,
            subscription=self.subscription,
            fail_if_not_exists=self.fail_if_not_exists,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        self.log.info("Deleted subscription %s", self.subscription)


class PubSubPublishMessageOperator(GoogleCloudBaseOperator):
    """
    Publish messages to a PubSub topic.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:PubSubPublishMessageOperator`

    Each Task publishes all provided messages to the same topic
    in a single Google Cloud project. If the topic does not exist, this
    task will fail. ::

        m1 = {"data": b"Hello, World!", "attributes": {"type": "greeting"}}
        m2 = {"data": b"Knock, knock"}
        m3 = {"attributes": {"foo": ""}}
        m4 = {"data": b"Who's there?", "attributes": {"ordering_key": "knock_knock"}}

        t1 = PubSubPublishMessageOperator(
            project_id="my-project",
            topic="my_topic",
            messages=[m1, m2, m3],
            create_topic=True,
            dag=dag,
        )

        t2 = PubSubPublishMessageOperator(
            project_id="my-project",
            topic="my_topic",
            messages=[m4],
            create_topic=True,
            enable_message_ordering=True,
            dag=dag,
        )

    ``project_id``, ``topic``, and ``messages`` are templated so you can use Jinja templating
    in their values.

    :param project_id: Optional, the Google Cloud project ID in which to work (templated).
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param topic: the topic to which to publish. Do not include the
        full topic path. In other words, instead of
        ``projects/{project}/topics/{topic}``, provide only
        ``{topic}``. (templated)
    :param messages: a list of messages to be published to the
        topic. Each message is a dict with one or more of the
        following keys-value mappings:
        * 'data': a bytestring (utf-8 encoded)
        * 'attributes': {'key1': 'value1', ...}
        Each message must contain at least a non-empty 'data' value
        or an attribute dict with at least one key (templated). See
        https://cloud.google.com/pubsub/docs/reference/rest/v1/PubsubMessage
    :param gcp_conn_id: The connection ID to use connecting to
        Google Cloud.
    :param enable_message_ordering: If true, messages published with the same
        ordering_key in PubsubMessage will be delivered to the subscribers in the order
        in which they are received by the Pub/Sub system. Otherwise, they may be
        delivered in any order. Default is False.
    :param enable_open_telemetry_tracing: If true, enables OpenTelemetry tracing for
        published messages. This allows distributed tracing of messages as they flow
        through Pub/Sub topics. Default is False.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "project_id",
        "topic",
        "messages",
        "enable_message_ordering",
        "enable_open_telemetry_tracing",
        "impersonation_chain",
    )
    ui_color = "#0273d4"

    def __init__(
        self,
        *,
        topic: str,
        messages: list,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        enable_message_ordering: bool = False,
        enable_open_telemetry_tracing: bool = False,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.topic = topic
        self.messages = messages
        self.gcp_conn_id = gcp_conn_id
        self.enable_message_ordering = enable_message_ordering
        self.enable_open_telemetry_tracing = enable_open_telemetry_tracing
        self.impersonation_chain = impersonation_chain

    @cached_property
    def pubsub_hook(self):
        return PubSubHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            enable_message_ordering=self.enable_message_ordering,
            enable_open_telemetry_tracing=self.enable_open_telemetry_tracing,
        )

    def execute(self, context: Context) -> None:
        self.log.info("Publishing to topic %s", self.topic)
        self.pubsub_hook.publish(project_id=self.project_id, topic=self.topic, messages=self.messages)
        self.log.info("Published to topic %s", self.topic)

    def get_openlineage_facets_on_complete(self, _) -> OperatorLineage:
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.openlineage.extractors import OperatorLineage

        project_id = self.project_id or self.pubsub_hook.project_id
        output_dataset = [Dataset(namespace="pubsub", name=f"topic:{project_id}:{self.topic}")]

        return OperatorLineage(outputs=output_dataset)


class PubSubPullOperator(GoogleCloudBaseOperator):
    """
    Pulls messages from a PubSub subscription and passes them through XCom.

    If the queue is empty, returns empty list - never waits for messages.
    If you do need to wait, please use :class:`airflow.providers.google.cloud.sensors.PubSubPullSensor`
    instead.

    .. seealso::
        For more information on how to use this operator and the PubSubPullSensor, take a look at the guide:
        :ref:`howto/operator:PubSubPullSensor`

    This operator will pull up to ``max_messages`` messages from the
    specified PubSub subscription. When the subscription returns messages,
    the messages will be returned immediately from the operator and passed through XCom for downstream tasks.

    If ``ack_messages`` is set to True, messages will be immediately
    acknowledged before being returned, otherwise, downstream tasks will be
    responsible for acknowledging them.

    ``project_id `` and ``subscription`` are templated so you can use Jinja templating in their values.

    :param project_id: the Google Cloud project ID for the subscription (templated)
    :param subscription: the Pub/Sub subscription name. Do not include the
        full subscription path.
    :param max_messages: The maximum number of messages to retrieve per
        PubSub pull request
    :param ack_messages: If True, each message will be acknowledged
        immediately rather than by any downstream tasks
    :param gcp_conn_id: The connection ID to use connecting to
        Google Cloud.
    :param messages_callback: (Optional) Callback to process received messages.
        Its return value will be saved to XCom.
        If you are pulling large messages, you probably want to provide a custom callback.
        If not provided, the default implementation will convert `ReceivedMessage` objects
        into JSON-serializable dicts using `google.protobuf.json_format.MessageToDict` function.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: If True, run the task in the deferrable mode.
    :param poll_interval: Time (seconds) to wait between two consecutive calls to check the job.
        The default is 300 seconds.
    """

    template_fields: Sequence[str] = (
        "project_id",
        "subscription",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        project_id: str,
        subscription: str,
        max_messages: int = 5,
        ack_messages: bool = False,
        messages_callback: Callable[[list[ReceivedMessage], Context], Any] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 300,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.gcp_conn_id = gcp_conn_id
        self.project_id = project_id
        self.subscription = subscription
        self.max_messages = max_messages
        self.ack_messages = ack_messages
        self.messages_callback = messages_callback
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    def execute(self, context: Context) -> list:
        if self.deferrable:
            self.defer(
                trigger=PubsubPullTrigger(
                    subscription=self.subscription,
                    project_id=self.project_id,
                    max_messages=self.max_messages,
                    ack_messages=self.ack_messages,
                    gcp_conn_id=self.gcp_conn_id,
                    poke_interval=self.poll_interval,
                    impersonation_chain=self.impersonation_chain,
                ),
                method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
            )
        hook = PubSubHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        pulled_messages = hook.pull(
            project_id=self.project_id,
            subscription=self.subscription,
            max_messages=self.max_messages,
            return_immediately=True,
        )

        handle_messages = self.messages_callback or self._default_message_callback

        ret = handle_messages(pulled_messages, context)

        if pulled_messages and self.ack_messages:
            hook.acknowledge(
                project_id=self.project_id,
                subscription=self.subscription,
                messages=pulled_messages,
            )

        return ret

    def execute_complete(self, context: Context, event: dict[str, Any]):
        """If messages_callback is provided, execute it; otherwise, return immediately with trigger event message."""
        if event["status"] == "success":
            self.log.info("Sensor pulls messages: %s", event["message"])
            messages_callback = self.messages_callback or self._default_message_callback
            _return_value = messages_callback(event["message"], context)
            return _return_value

        self.log.info("Sensor failed: %s", event["message"])
        raise AirflowException(event["message"])

    def _default_message_callback(
        self,
        pulled_messages: list[ReceivedMessage],
        context: Context,
    ) -> list:
        """
        Act as a default message callback.

        This method can be overridden by subclasses or by `messages_callback` constructor argument.

        This default implementation converts `ReceivedMessage` objects into JSON-serializable dicts.

        :param pulled_messages: messages received from the topic.
        :param context: same as in `execute`
        :return: value to be saved to XCom.
        """
        messages_json = [ReceivedMessage.to_dict(m) for m in pulled_messages]

        return messages_json

    def get_openlineage_facets_on_complete(self, _) -> OperatorLineage:
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.openlineage.extractors import OperatorLineage

        output_dataset = [
            Dataset(namespace="pubsub", name=f"subscription:{self.project_id}:{self.subscription}")
        ]

        return OperatorLineage(outputs=output_dataset)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Spanner Hook."""

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, NamedTuple

from google.api_core.exceptions import AlreadyExists, GoogleAPICallError
from google.cloud.spanner_v1.client import Client
from sqlalchemy import create_engine

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.common.sql.hooks.sql import DbApiHook
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook, get_field
from airflow.providers.openlineage.sqlparser import DatabaseInfo

if TYPE_CHECKING:
    from google.cloud.spanner_v1.database import Database
    from google.cloud.spanner_v1.instance import Instance
    from google.cloud.spanner_v1.transaction import Transaction
    from google.longrunning.operations_grpc_pb2 import Operation

    from airflow.models.connection import Connection


class SpannerConnectionParams(NamedTuple):
    """Information about Google Spanner connection parameters."""

    project_id: str | None
    instance_id: str | None
    database_id: str | None


class SpannerHook(GoogleBaseHook, DbApiHook):
    """
    Hook for Google Cloud Spanner APIs.

    All the methods in the hook where project_id is used must be called with
    keyword arguments rather than positional.
    """

    conn_name_attr = "gcp_conn_id"
    default_conn_name = "google_cloud_spanner_default"
    conn_type = "gcpspanner"
    hook_name = "Google Cloud Spanner"

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )
        self._client: Client | None = None

    def _get_client(self, project_id: str) -> Client:
        """
        Provide a client for interacting with the Cloud Spanner API.

        :param project_id: The ID of the Google Cloud project.
        :return: Client
        """
        if not self._client:
            self._client = Client(
                project=project_id, credentials=self.get_credentials(), client_info=CLIENT_INFO
            )
        return self._client

    def _get_conn_params(self) -> SpannerConnectionParams:
        """Extract spanner database connection parameters."""
        extras = self.get_connection(self.gcp_conn_id).extra_dejson
        project_id = get_field(extras, "project_id") or self.project_id
        instance_id = get_field(extras, "instance_id")
        database_id = get_field(extras, "database_id")
        return SpannerConnectionParams(project_id, instance_id, database_id)

    def get_uri(self) -> str:
        """Override DbApiHook get_uri method for get_sqlalchemy_engine()."""
        project_id, instance_id, database_id = self._get_conn_params()
        if not all([instance_id, database_id]):
            raise AirflowException("The instance_id or database_id were not specified")
        return f"spanner+spanner:///projects/{project_id}/instances/{instance_id}/databases/{database_id}"

    def get_sqlalchemy_engine(self, engine_kwargs=None):
        """
        Get an sqlalchemy_engine object.

        :param engine_kwargs: Kwargs used in :func:`~sqlalchemy.create_engine`.
        :return: the created engine.
        """
        if engine_kwargs is None:
            engine_kwargs = {}
        project_id, _, _ = self._get_conn_params()
        spanner_client = self._get_client(project_id=project_id)
        return create_engine(self.get_uri(), connect_args={"client": spanner_client}, **engine_kwargs)

    @GoogleBaseHook.fallback_to_default_project_id
    def get_instance(
        self,
        instance_id: str,
        project_id: str,
    ) -> Instance | None:
        """
        Get information about a particular instance.

        :param project_id: Optional, The ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :param instance_id: The ID of the Cloud Spanner instance.
        :return: Spanner instance
        """
        instance = self._get_client(project_id=project_id).instance(instance_id=instance_id)
        if not instance.exists():
            return None
        return instance

    def _apply_to_instance(
        self,
        project_id: str,
        instance_id: str,
        configuration_name: str,
        node_count: int,
        display_name: str,
        func: Callable[[Instance], Operation],
    ) -> None:
        """
        Invoke a method on a given instance by applying a specified Callable.

        :param project_id: The ID of the Google Cloud project that owns the Cloud Spanner database.
        :param instance_id: The ID of the instance.
        :param configuration_name: Name of the instance configuration defining how the
            instance will be created. Required for instances which do not yet exist.
        :param node_count: (Optional) Number of nodes allocated to the instance.
        :param display_name: (Optional) The display name for the instance in the Cloud
            Console UI. (Must be between 4 and 30 characters.) If this value is not set
            in the constructor, will fall back to the instance ID.
        :param func: Method of the instance to be called.
        """
        instance = self._get_client(project_id=project_id).instance(
            instance_id=instance_id,
            configuration_name=configuration_name,
            node_count=node_count,
            display_name=display_name,
        )
        try:
            operation: Operation = func(instance)
        except GoogleAPICallError as e:
            self.log.error("An error occurred: %s. Exiting.", e.message)
            raise e

        if operation:
            result = operation.result()
            self.log.info(result)

    @GoogleBaseHook.fallback_to_default_project_id
    def create_instance(
        self,
        instance_id: str,
        configuration_name: str,
        node_count: int,
        display_name: str,
        project_id: str,
    ) -> None:
        """
        Create a new Cloud Spanner instance.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param configuration_name: The name of the instance configuration defining how the
            instance will be created. Possible configuration values can be retrieved via
            https://cloud.google.com/spanner/docs/reference/rest/v1/projects.instanceConfigs/list
        :param node_count: (Optional) The number of nodes allocated to the Cloud Spanner
            instance.
        :param display_name: (Optional) The display name for the instance in the Google Cloud Console.
            Must be between 4 and 30 characters. If this value is not passed, the name falls back
            to the instance ID.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: None
        """
        self._apply_to_instance(
            project_id, instance_id, configuration_name, node_count, display_name, lambda x: x.create()
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def update_instance(
        self,
        instance_id: str,
        configuration_name: str,
        node_count: int,
        display_name: str,
        project_id: str,
    ) -> None:
        """
        Update an existing Cloud Spanner instance.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param configuration_name: The name of the instance configuration defining how the
            instance will be created. Possible configuration values can be retrieved via
            https://cloud.google.com/spanner/docs/reference/rest/v1/projects.instanceConfigs/list
        :param node_count: (Optional) The number of nodes allocated to the Cloud Spanner
            instance.
        :param display_name: (Optional) The display name for the instance in the Google Cloud
            Console. Must be between 4 and 30 characters. If this value is not set in
            the constructor, the name falls back to the instance ID.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: None
        """
        self._apply_to_instance(
            project_id, instance_id, configuration_name, node_count, display_name, lambda x: x.update()
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_instance(self, instance_id: str, project_id: str) -> None:
        """
        Delete an existing Cloud Spanner instance.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: None
        """
        instance = self._get_client(project_id=project_id).instance(instance_id)
        try:
            instance.delete()
            return
        except GoogleAPICallError as e:
            self.log.error("An error occurred: %s. Exiting.", e.message)
            raise e

    @GoogleBaseHook.fallback_to_default_project_id
    def get_database(
        self,
        instance_id: str,
        database_id: str,
        project_id: str,
    ) -> Database | None:
        """
        Retrieve a database in Cloud Spanner; return None if the database does not exist in the instance.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param database_id: The ID of the database in Cloud Spanner.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: Database object or None if database does not exist
        """
        instance = self._get_client(project_id=project_id).instance(instance_id=instance_id)
        if not instance.exists():
            raise AirflowException(f"The instance {instance_id} does not exist in project {project_id} !")
        database = instance.database(database_id=database_id)
        if not database.exists():
            return None

        return database

    @GoogleBaseHook.fallback_to_default_project_id
    def create_database(
        self,
        instance_id: str,
        database_id: str,
        ddl_statements: list[str],
        project_id: str,
    ) -> None:
        """
        Create a new database in Cloud Spanner.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param database_id: The ID of the database to create in Cloud Spanner.
        :param ddl_statements: The string list containing DDL for the new database.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: None
        """
        instance = self._get_client(project_id=project_id).instance(instance_id=instance_id)
        if not instance.exists():
            raise AirflowException(f"The instance {instance_id} does not exist in project {project_id} !")
        database = instance.database(database_id=database_id, ddl_statements=ddl_statements)
        try:
            operation: Operation = database.create()
        except GoogleAPICallError as e:
            self.log.error("An error occurred: %s. Exiting.", e.message)
            raise e

        if operation:
            result = operation.result()
            self.log.info(result)

    @GoogleBaseHook.fallback_to_default_project_id
    def update_database(
        self,
        instance_id: str,
        database_id: str,
        ddl_statements: list[str],
        project_id: str,
        operation_id: str | None = None,
    ) -> None:
        """
        Update DDL of a database in Cloud Spanner.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param database_id: The ID of the database in Cloud Spanner.
        :param ddl_statements: The string list containing DDL for the new database.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :param operation_id: (Optional) The unique per database operation ID that can be
            specified to implement idempotency check.
        :return: None
        """
        instance = self._get_client(project_id=project_id).instance(instance_id=instance_id)
        if not instance.exists():
            raise AirflowException(f"The instance {instance_id} does not exist in project {project_id} !")
        database = instance.database(database_id=database_id)
        try:
            operation = database.update_ddl(ddl_statements=ddl_statements, operation_id=operation_id)
            if operation:
                result = operation.result()
                self.log.info(result)
            return
        except AlreadyExists as e:
            if e.code == 409 and operation_id in e.message:
                self.log.info(
                    "Replayed update_ddl message - the operation id %s was already done before.",
                    operation_id,
                )
                return
        except GoogleAPICallError as e:
            self.log.error("An error occurred: %s. Exiting.", e.message)
            raise e

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_database(self, instance_id: str, database_id, project_id: str) -> bool:
        """
        Drop a database in Cloud Spanner.

        :param instance_id: The ID of the Cloud Spanner instance.
        :param database_id: The ID of the database in Cloud Spanner.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: True if everything succeeded
        """
        instance = self._get_client(project_id=project_id).instance(instance_id=instance_id)
        if not instance.exists():
            raise AirflowException(f"The instance {instance_id} does not exist in project {project_id} !")
        database = instance.database(database_id=database_id)
        if not database.exists():
            self.log.info(
                "The database %s is already deleted from instance %s. Exiting.", database_id, instance_id
            )
            return False
        try:
            database.drop()
        except GoogleAPICallError as e:
            self.log.error("An error occurred: %s. Exiting.", e.message)
            raise e

        return True

    @GoogleBaseHook.fallback_to_default_project_id
    def execute_dml(
        self,
        instance_id: str,
        database_id: str,
        queries: list[str],
        project_id: str,
    ) -> list[int]:
        """
        Execute an arbitrary DML query (INSERT, UPDATE, DELETE).

        :param instance_id: The ID of the Cloud Spanner instance.
        :param database_id: The ID of the database in Cloud Spanner.
        :param queries: The queries to execute.
        :param project_id: Optional, the ID of the Google Cloud project that owns the Cloud Spanner
            database. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
        :return: list of numbers of affected rows by DML query
        """
        db = (
            self._get_client(project_id=project_id)
            .instance(instance_id=instance_id)
            .database(database_id=database_id)
        )

        def _tx_runner(tx: Transaction) -> dict[str, int]:
            return self._execute_sql_in_transaction(tx, queries)

        result = db.run_in_transaction(_tx_runner)

        result_rows_count_per_query = []
        for i, (sql, rc) in enumerate(result.items(), start=1):
            if not sql.startswith("SELECT"):
                preview = sql if len(sql) <= 300 else sql[:300] + "…"
                self.log.info("[DML %d/%d] affected rows=%d | %s", i, len(result), rc, preview)
                result_rows_count_per_query.append(rc)
        return result_rows_count_per_query

    @staticmethod
    def _execute_sql_in_transaction(transaction: Transaction, queries: list[str]) -> dict[str, int]:
        counts: OrderedDict[str, int] = OrderedDict()
        for sql in queries:
            rc = transaction.execute_update(sql)
            counts[sql] = rc
        return counts

    def _get_openlineage_authority_part(self, connection: Connection) -> str | None:
        """Build Spanner-specific authority part for OpenLineage. Returns {project}/{instance}."""
        extras = connection.extra_dejson
        project_id = extras.get("project_id")
        instance_id = extras.get("instance_id")

        if not project_id or not instance_id:
            return None

        return f"{project_id}/{instance_id}"

    def get_openlineage_database_dialect(self, connection: Connection) -> str:
        """Return database dialect for OpenLineage."""
        return "spanner"

    def get_openlineage_database_info(self, connection: Connection) -> DatabaseInfo:
        """Return Spanner specific information for OpenLineage."""
        extras = connection.extra_dejson
        database_id = extras.get("database_id")

        return DatabaseInfo(
            scheme=self.get_openlineage_database_dialect(connection),
            authority=self._get_openlineage_authority_part(connection),
            database=database_id,
            information_schema_columns=[
                "table_schema",
                "table_name",
                "column_name",
                "ordinal_position",
                "spanner_type",
            ],
        )

    def get_openlineage_default_schema(self) -> str | None:
        """
        Spanner expose 'public' or '' schema depending on dialect(Postgres vs GoogleSQL).

        SQLAlchemy dialect for Spanner does not expose default schema, so we return None
        to follow the same approach.
        """
        return None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that show how to use various Looker
operators to submit PDT materialization job and manage it.
"""

from __future__ import annotations

from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.looker import LookerStartPdtBuildOperator
from airflow.providers.google.cloud.sensors.looker import LookerCheckPdtBuildSensor

DAG_ID = "gcp_looker"
LOOKER_CONNECTION_ID = "your_airflow_connection_for_looker"
LOOKER_MODEL = "your_lookml_model"
LOOKER_VIEW = "your_lookml_view"

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "looker"],
) as dag:
    # [START cloud_looker_async_start_pdt_sensor]
    start_pdt_task_async = LookerStartPdtBuildOperator(
        task_id="start_pdt_task_async",
        looker_conn_id=LOOKER_CONNECTION_ID,
        model=LOOKER_MODEL,
        view=LOOKER_VIEW,
        asynchronous=True,
    )

    check_pdt_task_async_sensor = LookerCheckPdtBuildSensor(
        task_id="check_pdt_task_async_sensor",
        looker_conn_id=LOOKER_CONNECTION_ID,
        materialization_id=start_pdt_task_async.output,
        poke_interval=10,
    )
    # [END cloud_looker_async_start_pdt_sensor]

    # [START how_to_cloud_looker_start_pdt_build_operator]
    build_pdt_task = LookerStartPdtBuildOperator(
        task_id="build_pdt_task",
        looker_conn_id=LOOKER_CONNECTION_ID,
        model=LOOKER_MODEL,
        view=LOOKER_VIEW,
    )
    # [END how_to_cloud_looker_start_pdt_build_operator]

    start_pdt_task_async >> check_pdt_task_async_sensor

    build_pdt_task

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
